//Chi2 => program to compute chi^2 on binned sharp_combine output.
//Nov 28 2007
//Author: Michael Attard  

//Layout of <NAMES.LIST>:
//b1.fits
//b2.fits
//bt.fits
//where bt.fits is the net cumulative file, and b1, b2, ... etc are the cumulative files for each bin respectively

//Version 2 - now allows for input data files to be generated with sharp_combine_v5 and older versions.

//Version 2.1 - now crops portions of the chi-squared map for localized analysis. Format for <POINTS.LIST> is:

//x y r => where x and y is the point in pixels (graphs taken to start at 1,1 not 0,0) and r is the radius of 
//the crop

//Version 2.2 - now allows a user to update the error maps from sharp_combine output.  Use '-apply' flag.
//            - user can also print out an image of I/sig_I by calling the '-debug' flag.
//            - eliminate '-a' flag, now read in image dimensions directly from sharp_combine output 

//Version 2.3 - now allows for a DC offset to be calculated and removed from the reduced chi-squared result.
//            - two flags can be used to this end, '-dc' and '-adc';
//            - '-dc' => computes the DC offset from a circular patch of the image with a radius of 4 pixels.
//            - '-adc' => computes the DC offset from an annulus with an inner radius of r1 and an outer radius of r2.

//Version 2.4 - now computes the inflation factor: sigma_p'/sigma_p.  For use with polsharp's 'maxsig' flag.

//To allow NAN to work on zamin
#define _GNU_SOURCE

//Libaries
#include <stdio.h>
#include <math.h>
#include "fitsio.h"
#include <stdlib.h>
#include <string.h>
#include "longnam.h"

//Def'n
#define Min -10E-20
#define Max 10E-20

//To allow NAN to work on zamin
#if defined (NAN)
      #define BLANKING  NAN
#else
      #define BLANKING  FP_NAN
#endelse
#endif

//Functions
void Abort(char *msg, int status);
void read_command_line(int argc, char *argv[], double *r1, double *r2);
void read_info(double *stokes, double *sig_stokes,int bin);
void countbin(int *b);
double** dfree2D(double **P,int x);
void read_bin(double *image, int axis_x, int axis_y, char stokes, char *Names, double *m, double r1, double r2);
void read_bin2(double *image, int axis_x, int axis_y, char stokes, char *Names, double *m, double r1, double r2);
void unreduce(double *inimage, double *I, int axis_x, int axis_y);
void unreduce2(double *inimage, double *I, double *sI, double *qu, int axis_x, int axis_y);
void X(int b, int axis_x, int axis_y, double *X, double *M, double **Si, double *Sf, double **sS, double **m, int n);
void print_image(int axis_x, int axis_y, double *XI, double *XQ, double *XU, double *MI, double *MQ, double *MU, char *outname);
void debugging(int b);
void print_image_debug(int bin, double *I, double *Q, double *U, double *sI, double *sQ, double *sU, int dimx, int dimy, char *outname);
double** allocate2D(int x, int y);
void set_to_zero(double *A, int N);
void avevar(double data[], double *M,int n, double *ave, double *var);
double** readPoints(int *N);
void printPoints(double **P, int N);
void crop(double **P, int N, double *XI, double *XQ, double *XU, double *MI, double *MQ, double *MU, int axis_x, int axis_y);
void correct_error_maps(char *file, double *sI, double *sQ, double *sU, double *XI, double *XQ, double *XU, double *MI, double *MQ, double *MU, int axis_x, int axis_y);
void I_over_sigI_ratio(int axis_x, int axis_y, double *I, double *sQ);
void check_dimensions(int *axis_x, int *axis_y, char *Names);
void compute_dc(double **m, int b);
void compute_maxsig_factor(double *I, double *Q, double *U, double *sI, double *sQ, double *sU, double *XI, double *XQ, double *XU, double *M, int axis_x, int axis_y, double **P, int N);

//Global Vars
char infile[FILENAME_MAX],cfile[FILENAME_MAX], debug='n', v4='n',points='n',update='n',dc='n',dc2='n';

//main function
int main (int argc, char *argv[]) {
	
	//variable declaration
	int b=0,i=0,axis_x=0,axis_y=0, N=0;
	char *Names;//Array with the structure: Names[Filename size] 
	double **Ii, **Qi, **Ui, **sIi, **sQi, **sUi, *MI, *MQ, *MU, *XI, *XQ, *XU, ave, var, **P, **m, r1, r2;
	FILE *fp;
	
	//read command line
	read_command_line(argc,argv,&r1,&r2);
	
	//count # of bins
	countbin(&b);
	
	//read in axis_x and axis_y
	Names = calloc(FILENAME_MAX,sizeof(char));
	fp = fopen(infile,"r");
	fscanf(fp,"%s",Names);
	check_dimensions(&axis_x,&axis_y,Names);
	rewind(fp);
	printf("Image Dimensions of 1st file (x,y): (%d,%d)\n",axis_x,axis_y);
	if(dc2=='y')
	  printf("Annulus parameters for DC offset calculation: inner radius = %lf, outer radius = %lf\n",r1,r2);
	
	//allocate memory
	m = allocate2D(b,3);
    Ii = allocate2D(b,axis_x*axis_y);
	Qi = allocate2D(b,axis_x*axis_y);
	Ui = allocate2D(b,axis_x*axis_y);
	sIi = allocate2D(b,axis_x*axis_y);
	sQi = allocate2D(b,axis_x*axis_y);
	sUi = allocate2D(b,axis_x*axis_y);
	MI = calloc(axis_x*axis_y,sizeof(double));
	MQ = calloc(axis_x*axis_y,sizeof(double));
	MU = calloc(axis_x*axis_y,sizeof(double));
	XI = calloc(axis_x*axis_y,sizeof(double));
	XQ = calloc(axis_x*axis_y,sizeof(double));
	XU = calloc(axis_x*axis_y,sizeof(double));

	//execute debugging procedure if requested
	if(debug=='y')
	  debugging(b);
		
	//main loop to read and process each bin of data
	while(fscanf(fp,"%s",Names)!=EOF){
    
	   //check for proper dimensions
	   check_dimensions(&axis_x,&axis_y,Names);
	
	   if(v4=='y'){	
	     read_bin(Ii[i],axis_x,axis_y,'i',Names,m[i],r1,r2);
	     read_bin(Qi[i],axis_x,axis_y,'q',Names,m[i],r1,r2);
         read_bin(Ui[i],axis_x,axis_y,'u',Names,m[i],r1,r2);
         read_bin(sIi[i],axis_x,axis_y,'x',Names,m[i],r1,r2);
         read_bin(sQi[i],axis_x,axis_y,'y',Names,m[i],r1,r2);
         read_bin(sUi[i],axis_x,axis_y,'z',Names,m[i],r1,r2);
	   }
	   else {
	     read_bin2(Ii[i],axis_x,axis_y,'i',Names,m[i],r1,r2);
	     read_bin2(Qi[i],axis_x,axis_y,'q',Names,m[i],r1,r2);
         read_bin2(Ui[i],axis_x,axis_y,'u',Names,m[i],r1,r2);
         read_bin2(sIi[i],axis_x,axis_y,'x',Names,m[i],r1,r2);
         read_bin2(sQi[i],axis_x,axis_y,'y',Names,m[i],r1,r2);
         read_bin2(sUi[i],axis_x,axis_y,'z',Names,m[i],r1,r2);
	   }
	   
	  //print out images if desired
	  if(debug=='y')
	    print_image_debug(i,Ii[i],Qi[i],Ui[i],sIi[i],sQi[i],sUi[i],axis_x,axis_y,"Bin");
	
	  //increment counter
	  i++;
	
	}
	
	//close file
	fclose(fp);
	
	//compute dc offset if requested
	if(dc=='y' || dc2=='y')
	  compute_dc(m,b);
	
	//calculate chi-squared maps
	X(b,axis_x,axis_y,XI,MI,Ii,Ii[b-1],sIi,m,0);
	X(b,axis_x,axis_y,XQ,MQ,Qi,Qi[b-1],sQi,m,1);
    X(b,axis_x,axis_y,XU,MU,Ui,Ui[b-1],sUi,m,2);
	
	//if cropping is requested
	if(points=='y'){
	  P = readPoints(&N);
	  if(debug=='y')
	    printPoints(P,N);
	  crop(P,N,XI,XQ,XU,MI,MQ,MU,axis_x,axis_y);
	}
	
	//compute mean's and Stdev's
	printf("Summary of results for whole map:\n");
	avevar(XI,MI,axis_x*axis_y,&ave,&var);
	printf("Reduced Chi Squared mean and standard dev. for the I map: %lf, %lf\n",ave,sqrt(var));
	avevar(XQ,MQ,axis_x*axis_y,&ave,&var);
	printf("Reduced Chi Squared mean and standard dev. for the Q map: %lf, %lf\n",ave,sqrt(var));
    avevar(XU,MU,axis_x*axis_y,&ave,&var);
	printf("Reduced Chi Squared mean and standard dev. for the U map: %lf, %lf\n",ave,sqrt(var));

    //compute inflation factor
    compute_maxsig_factor(Ii[b-1],Qi[b-1],Ui[b-1],sIi[b-1],sQi[b-1],sUi[b-1],XI,XQ,XU,MI,axis_x,axis_y,P,N);

	//print to output file 'chi2.fits'
	print_image(axis_x,axis_y,XI,XQ,XU,MI,MQ,MU,"chi2");
	
	//correct error maps in the cumulative bin if requested
	if(update=='y')
	  correct_error_maps(Names,sIi[b-1],sQi[b-1],sUi[b-1],XI,XQ,XU,MI,MQ,MU,axis_x,axis_y);
	  
	//print out ratio.fits (a file showing the ratio of I/sigma_I) if requested
	if(debug=='y')
	  I_over_sigI_ratio(axis_x,axis_y,Ii[b-1],sIi[b-1]);  
	
	//free memory
	if(points=='y')
	  P = dfree2D(P,N);
	m = dfree2D(m,b);  
	Ii = dfree2D(Ii, b);
	Qi = dfree2D(Qi, b);
	Ui = dfree2D(Ui, b);
	sIi = dfree2D(sIi, b);
	sQi = dfree2D(sQi, b);
	sUi = dfree2D(sUi, b);
	free(MI);
	free(MQ);
    free(MU);
	free(XI);
	free(XQ);
	free(XU);
	free(Names);
	
	return 0;
}

//Prints out an image of whatever *I points to
void print_image_debug(int bin, double *I, double *Q, double *U, double *sI, double *sQ, double *sU, int dimx, int dimy, char *outname){

  fitsfile *out;
  char name[FILENAME_MAX]="",str[10];
  long naxes[2] = {dimx,dimy};
  int size = dimx*dimy,status=0;
  
  //Generate output file name
  strcat(name,"!");
  strcat(name,outname);
  sprintf(str,"%d",bin+1);
  strcat(name,str);
  strcat(name,".fits");
  printf("Printing: %s\n",name);

  //Create file
  if(fits_create_file(&out,name,&status))
    Abort("FITS error in generating output file",status);
  //Print to file	
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","I","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);	
  if(fits_write_img(out,TDOUBLE,1,size,I,&status)) 
	Abort("FITS error in writing image",status);
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","(Error-I)^2","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);	
  if(fits_write_img(out,TDOUBLE,1,size,sI,&status)) 
	Abort("FITS error in writing image",status);	
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","Q","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);	
  if(fits_write_img(out,TDOUBLE,1,size,Q,&status)) 
	Abort("FITS error in writing image",status);
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","(Error-Q)^2","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);	
  if(fits_write_img(out,TDOUBLE,1,size,sQ,&status)) 
	Abort("FITS error in writing image",status);	
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","U","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);	
  if(fits_write_img(out,TDOUBLE,1,size,U,&status)) 
	Abort("FITS error in writing image",status);	
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","(Error-U)^2","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);	
  if(fits_write_img(out,TDOUBLE,1,size,sU,&status)) 
	Abort("FITS error in writing image",status);	
	
  //Close file	
  if(fits_close_file(out,&status))
    Abort("FITS error in closing output file",status);
}

//Function used for premature exit
void Abort(char *msg, int status)
{
  char *usage ="Use:chi2 -f <NAMES.LIST> {-v4 -debug -c <POINTS.LIST> -update -dc -adc <r1> <r2>}.\n\n FLAGS:\n     -f: The file <NAMES.LIST> is to contain a list of all the binned data. The file names should be entered sequentially in a single 'column'.\n     -v4: The '-v4' flag should be present if the binned data is processed by sharp_combine v4.xx or earlier.\n     -c: The '-c' flag is followed by the file <POINTS.LIST> that should indicate locations for cropping the chi^2 maps.\n     -update: Correct the I, Q, and U error maps in the cumulative bin image stored in <NAMES.LIST>.\n              Note this flag will only work if the '-v4' flag is not present.\n     -debug: The '-debug' flag should be present if the user wants additional information to be generated by the program. This includes:\n             *print out of the unreduced Stokes parameters & there associated variances into 'Binx.fits' files (where x=1 to no. of bins)\n             *print out of the # of bins read into the program\n             *if '-c' flag is called, chi2.c will print out the contents of <POINTS.LIST> to the screen\n             *print out an image of I / sigma_I for the cumulative bin image.  This ratio is stored in 'ratio.fits'\n     -dc: Will prompt chi2.c to compute a DC offset and remove this component from the reduced-chi-squared result.\n          This DC offset is computed from data enclosed within a radius of 4 pixels from the image center.\n     -adc: Will prompt chi2.c to compute a DC offset from an annulus specified by the user.  This offset will be removed from the results.\n           The annulus is specified via the radial parameters <r1> & <r2>, where <r1> < <r2>.\n\n  *Note that the parameters within { } are not necessary for the proper execution of the program.\n  *Also note that for more info see 'chi2README.rtf' in the /bin directory on sharp@zamin.";
  
  fprintf(stderr,"Error => %s\n",msg);
  fprintf(stderr,"%s\n",usage);
  fprintf(stderr,"Program Terminated, status: %d\n",status);
  exit(1);
}

//Function used for obtaining all information passed from the command line
void read_command_line(int argc, char *argv[], double *r1, double *r2)
{
  int arg = 1;
  char *ver = "chi2 version v2.4"; 
  printf("%s\n",ver);
  //initalize file name variables
  strcpy(infile,"");
  strcpy(cfile,"");
  if(argc>=3){
	while(arg<argc){
	  
	  //check if input file is listed	 
	  if(strncmp(argv[arg],"-f",2)==0){
	    arg++;
		if(arg!=argc)
		  sscanf(argv[arg],"%s",infile);
		else
		  Abort("No 'NAMES.LIST' file name entered",0);
	  }//end of if
	  
	  //check if cropping file is listed	 
	  if(strncmp(argv[arg],"-c",2)==0){
	    points='y';
	    arg++;
		if(arg!=argc)
		  sscanf(argv[arg],"%s",cfile);
		else
		  Abort("No 'POINTS.LIST' file name entered",0);
	  }//end of if
	  
	  if(strncmp(argv[arg],"-debug",6)==0)//check if debugging is requested
	    debug='y';
	
	  if(strncmp(argv[arg],"-v4",3)==0)//check if input is processed by sharp_combine v4 or earlier is requested
	    v4='y';	
	
	  if(strncmp(argv[arg],"-update",7)==0)//check if error modification is requested
	    update='y';			
		
	  if(strncmp(argv[arg],"-dc",3)==0)//check if DC offset calculation is requested
	    dc='y';					
		
	  //check if annulus DC offset option is requested	 
	  if(strncmp(argv[arg],"-adc",4)==0){
	    dc2='y';
	    arg++;
		if(arg!=argc)
		  sscanf(argv[arg],"%lf",r1);
		else
		  Abort("No <r1> parameter entered along with -adc flag",0);
		arg++;
		if(arg!=argc)
		  sscanf(argv[arg],"%lf",r2);
		else
		  Abort("No <r2> parameter entered along with -adc flag",0);  
		if((*r1)>(*r2))
		  Abort("Error: r1 > r2, supposed to be r1 < r2",0);  
	  }//end of if	
										
	  arg++;
    }//end of while
  }//end of if 
  else
	Abort("Incorrect number of parameters at command line",0);
	
  //check if -v4 and -update flags were both requested (a no-no)	
  if(v4=='y' && update=='y')
    Abort("flags -v4 and -update both requested",0);	
	
  //check to see that -dc and -dc2 were not both selected
  if(dc=='y' && dc2=='y')
    Abort("flags -dc and -dc2 both requested",0);	
	
}//end of function

//Function that reads in information from <NAMES.LIST>.
void read_bin(double *image, int axis_x, int axis_y, char stokes, char *Names, double *m, double r1, double r2)
{
  //var dec's
  FILE *fp;
  fitsfile *in;
  int status=0,null2,i,j;
  double *temp_image, *I, *sI, null = DOUBLENULLVALUE,r,count=0;
  //open <NAMES.LIST>
  fp = fopen(infile,"r");

  //allocate memory
  temp_image = calloc(axis_x*axis_y,sizeof(double));
  I = calloc(axis_x*axis_y,sizeof(double));
  sI = calloc(axis_x*axis_y,sizeof(double));
  		  
  //read in data from FITS file
  printf("Reading in %s for processing\n",Names);
  if(fits_open_file(&in,Names,READWRITE,&status))
    Abort("FITS error with opening SOLVE.FITS",status);
			  
  //check for the appropriate HDU to read in
  if(strncmp(&stokes,"i",1)==0){//read in I map
	if(fits_movabs_hdu(in,2,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading I image from combine.fits",status);
	//set to zero
	set_to_zero(image,axis_x*axis_y);		
	if(dc=='y'){//compute mean over circular patch 8pix in diameter
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r<=4 && image[j*axis_x + i]!=0){//check to see if we're within the circular patch
		    m[0] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[0] /= count;
	}//end if
	if(dc2=='y'){//compute mean over annulus
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r <= r2 && r >= r1 && image[j*axis_x + i]!=0){//check to see if we're within the annular patch
		    m[0] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[0] /= count;
	}//end if
  
  }
			
  if(strncmp(&stokes,"x",1)==0){//read in error I map
	if(fits_movabs_hdu(in,3,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading error I image from combine.fits",status);
	//set to zero
	set_to_zero(image,axis_x*axis_y);	
	//square error map (generate variance map)
	for(i=0;i<axis_x*axis_y;i++)
	  image[i] = pow(image[i],2);
  }

  if(strncmp(&stokes,"q",1)==0){//read in q map
	if(fits_movabs_hdu(in,8,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,I,&null2,&status))
	  Abort("FITS error with reading I image from combine.fits",status);
	if(fits_movabs_hdu(in,4,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading q image from combine.fits",status);
	//set to zero
	set_to_zero(I,axis_x*axis_y);	
	set_to_zero(image,axis_x*axis_y);
	//compute Q = q*I
	unreduce(image,I,axis_x,axis_y);
	if(dc=='y'){//compute mean over circular patch 8pix in diameter
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r<=4 && image[j*axis_x + i]!=0){//check to see if we're within the circular patch
		    m[1] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[1] /= count;
	}//end if
	if(dc2=='y'){//compute mean over annulus
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r <= r2 && r >= r1 && image[j*axis_x + i]!=0){//check to see if we're within the annular patch
		    m[1] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[1] /= count;
	}//end if
		
  }

  if(strncmp(&stokes,"y",1)==0){//read in error q map
	if(fits_movabs_hdu(in,8,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,I,&null2,&status))
	  Abort("FITS error with reading I image from combine.fits",status);
	if(fits_movabs_hdu(in,3,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,sI,&null2,&status))
	  Abort("FITS error with reading error I image from combine.fits",status);	
	if(fits_movabs_hdu(in,4,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,temp_image,&null2,&status))
	  Abort("FITS error with reading q image from combine.fits",status);
	if(fits_movabs_hdu(in,5,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading error q image from combine.fits",status);	
	//set to zero
	set_to_zero(I,axis_x*axis_y);	
	set_to_zero(sI,axis_x*axis_y);
	set_to_zero(temp_image,axis_x*axis_y);
	set_to_zero(image,axis_x*axis_y);
	//compute sig-Q^2 = I^2 * sig-q^2 + q^2 * sig-I^2 
	unreduce2(image,I,sI,temp_image,axis_x,axis_y);
  }
			
  if(strncmp(&stokes,"u",1)==0){//read in u map
	if(fits_movabs_hdu(in,8,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,I,&null2,&status))
	  Abort("FITS error with reading I image from combine.fits",status);
	if(fits_movabs_hdu(in,6,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading u image from combine.fits",status);
	//set to zero
	set_to_zero(I,axis_x*axis_y);
	set_to_zero(image,axis_x*axis_y);				
	//compute U = u*I
	unreduce(image,I,axis_x,axis_y);
	if(dc=='y'){//compute mean over circular patch 8pix in diameter
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r<=4 && image[j*axis_x + i]!=0){//check to see if we're within the circular patch
		    m[2] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[2] /= count;
	}//end if
	if(dc2=='y'){//compute mean over annulus
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r <= r2 && r >= r1 && image[j*axis_x + i]!=0){//check to see if we're within the annular patch
		    m[2] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[2] /= count;
	}//end if
	
  }

  if(strncmp(&stokes,"z",1)==0){//read in error u map
	if(fits_movabs_hdu(in,8,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,I,&null2,&status))
	  Abort("FITS error with reading I image from combine.fits",status);
	if(fits_movabs_hdu(in,3,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,sI,&null2,&status))
	  Abort("FITS error with reading error I image from combine.fits",status);	
	if(fits_movabs_hdu(in,6,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,temp_image,&null2,&status))
	  Abort("FITS error with reading u image from combine.fits",status);
	if(fits_movabs_hdu(in,7,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading error u image from combine.fits",status);
	//set to zero
	set_to_zero(I,axis_x*axis_y);	
	set_to_zero(sI,axis_x*axis_y);
	set_to_zero(temp_image,axis_x*axis_y);
	set_to_zero(image,axis_x*axis_y);	
	//compute sig-U^2 = I^2 * sig-u^2 + u^2 * sig-I^2 
	unreduce2(image,I,sI,temp_image,axis_x,axis_y);	
  }
			
  //close FITS file
  if(fits_close_file(in,&status))
    Abort("FITS error in closing data file",status);	
		  
  //close <NAMES.LIST>
  fclose(fp);
  //free memory
  free(temp_image);
  free(I);
  free(sI);
  
}

//Function that counts how many bins are indicated in <NAMES.LIST>.
void countbin(int *b)
{
  //var dec's
  FILE *fp;
  char temp[FILENAME_MAX];
  //open <NAMES.LIST>
  fp = fopen(infile,"r");
  printf("Reading: %s\n",infile);

  //loop through <NAMES.LIST>, counting # of bins
  while(fscanf(fp,"%s",temp) != EOF){ 	
	(*b)++;
	strcpy(temp,"");	
  }
  
  //close <NAMES.LIST>
  fclose(fp);
  
}

//Function used for debugging, prints out data on <NAMES.LIST>
void debugging(int b){
  
  //print out # of bins
  printf("*** Total # of bins: %d***\n",b);

}

//Function used to free memory in 2D array P[x][y]
double** dfree2D(double **P,int x){

  int i;
  
  for(i=0;i<x;i++)
    free(P[i]);
	
  free(P);	
 
  return(P);

}

//Function that calculates outimage = inimage*I
void unreduce(double *inimage, double *I, int axis_x, int axis_y){

  int i,j;
  
  for(i=0;i<axis_x;i++)
    for(j=0;j<axis_y;j++)
	  inimage[axis_x*j+i] = inimage[axis_x*j+i]*I[axis_x*j+i];

}

//Function that computes inimage^2 = I^2 * inimage^2 + qu^2 * sI^2 
void unreduce2(double *inimage, double *I, double *sI, double *qu, int axis_x, int axis_y){

  int i,j;//,c=0,cc=0;
  //double aveI=0, avequ=0;
  
  //work out the <I^2> and <qu> quantities
  //for(i=0;i<axis_x*axis_y;i++){
    //if(I[i]!=0){
	  //c++;//count number of non-zero pixels
      //aveI += I[i]*I[i];
	//}
	//if(qu[i]!=0){
	  //cc++;//count non-zero pixels
	  //avequ += qu[i];
	//}
  //} 
  
  //aveI /= c;//average I^2
  //avequ /= cc;//average qu
  
  for(i=0;i<axis_x;i++)
    for(j=0;j<axis_y;j++)
	  inimage[axis_x*j+i] = pow(inimage[axis_x*j+i],2)*pow(I[axis_x*j+i],2) + pow(qu[axis_x*j+i],2)*pow(sI[axis_x*j+i],2);

}

//Function for calculating the chi-squared maps
void X(int b, int axis_x, int axis_y, double *X, double *M, double **Si, double *Sf, double **sS, double **m, int n){

  int i, j, x, y;
  
  //cycle through mask, assigning everything to 1
  for(y=0;y<axis_y;y++)
	for(x=0;x<axis_x;x++)
	  M[axis_x*y+x] = 1;    
  
  //if dc offset not requested, ensure all elements of **m are zero
  if(dc=='n' && dc2=='n'){
    for(i=0;i<b;i++)
      for(j=0;j<3;j++)
	    m[i][j] = 0;    
  }
  
  for(i=0;i<b-1;i++)
    for(y=0;y<axis_y;y++)
      for(x=0;x<axis_x;x++){
	    if(v4=='y')
          X[axis_x*y+x] += pow(Si[i][axis_x*y+x]-m[i][n]-Sf[axis_x*y+x],2)/sS[i][axis_x*y+x];
        else
		  X[axis_x*y+x] += pow(Si[i][axis_x*y+x]-m[i][n]-Sf[axis_x*y+x],2)/pow(sS[i][axis_x*y+x],2);
		  
		if(sS[i][axis_x*y+x]==0)
		  M[axis_x*y+x] = 0;
      }
  
  for(y=0;y<axis_y;y++)
	for(x=0;x<axis_x;x++)
	  X[axis_x*y+x] /= (b-2);  

}

//Function for printing the chi-squared images
void print_image(int axis_x, int axis_y, double *XI, double *XQ, double *XU, double *MI, double *MQ, double *MU, char *outname){

  fitsfile *out;
  char name[FILENAME_MAX]="";
  long naxes[2] = {axis_x,axis_y};
  int status=0;
  
  //Generate output file name
  strcat(name,"!");
  strcat(name,outname);
  strcat(name,".fits");
  printf("Printing: %s\n",name);

  //Create the file and print
  if(fits_create_file(&out,name,&status))
    Abort("FITS error in generating output file",status);
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","I","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);
  if(fits_write_img(out,TDOUBLE,1,axis_x*axis_y,XI,&status))
    Abort("FITS error in writing image",status);
	
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","Q","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);
  if(fits_write_img(out,TDOUBLE,1,axis_x*axis_y,XQ,&status))
    Abort("FITS error in writing image",status);

  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","U","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);
  if(fits_write_img(out,TDOUBLE,1,axis_x*axis_y,XU,&status))
    Abort("FITS error in writing image",status);

  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","I_Mask","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);
  if(fits_write_img(out,TDOUBLE,1,axis_x*axis_y,MI,&status))
    Abort("FITS error in writing image",status);

  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","Q_Mask","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);
  if(fits_write_img(out,TDOUBLE,1,axis_x*axis_y,MQ,&status))
    Abort("FITS error in writing image",status);
	
  if(fits_create_img(out,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(out,TSTRING,"EXTNAME","U_Mask","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);
  if(fits_write_img(out,TDOUBLE,1,axis_x*axis_y,MU,&status))
    Abort("FITS error in writing image",status);			
	
  if(fits_close_file(out,&status))
    Abort("FITS error in closing output file",status);

}

//Function used to allocate memory for 2D arrays, spits out P[x][y]
double** allocate2D(int x, int y){

  int i;
  double **P;
  
  P = calloc(x,sizeof(double));
  for(i=0;i<x;i++)
    P[i] = calloc(y,sizeof(double));
	
  return(P);
}

//Function that checks for spurrious numbers and sets them to zero
void set_to_zero(double *A, int N){

  int i;
  
  for(i=0;i<N;i++)
    if(A[i] <= Max && A[i] >= Min)
	  A[i] = 0;
	
}

//avevar function from NR in C, 2nd Edit (Pg. 617)
//modified to not process cells that = NAN
void avevar(double data[], double *M,int n, double *ave, double *var){
  /*Given array data[1..n], returns its mean as ave and its varience as var.*/
  unsigned long j;
  double s,ep;//do sqrt(var+ave^2)
  int c=0;

  for(*ave=0.0,j=0;j<n;j++){
    if(M[j] != 0){
	  *ave += data[j];
	  c++;
    }
  }
  
  *ave /= c;
  *var = ep = 0.0;
  
  for(j=0;j<n;j++){
    if(M[j] != 0){  
	  s=data[j]-(*ave);
      ep += s;
      *var += s*s;
	}
  }
  *var = (*var-ep*ep/c)/(c-1);
}

//Function that reads in information from <NAMES.LIST>.
void read_bin2(double *image, int axis_x, int axis_y, char stokes, char *Names, double *m, double r1, double r2)
{
  //var dec's
  FILE *fp;
  fitsfile *in;
  int status=0,null2,j,i;
  double null = DOUBLENULLVALUE,count,r;
  //open <NAMES.LIST>
  fp = fopen(infile,"r");
  		  
  //read in data from FITS file
  printf("Reading in %s for processing\n",Names);
  if(fits_open_file(&in,Names,READWRITE,&status))
    Abort("FITS error with opening SOLVE.FITS",status);
			  
  //check for the appropriate HDU to read in
  if(strncmp(&stokes,"i",1)==0){//read in I map
	if(fits_movabs_hdu(in,8,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading I image from combine.fits",status);
	//set to zero
	set_to_zero(image,axis_x*axis_y);
	if(dc=='y'){//compute mean over circular patch 8pix in diameter
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r<=4 && image[j*axis_x + i]!=0){//check to see if we're within the circular patch
		    m[0] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[0] /= count;
	}//end if	
	if(dc2=='y'){//compute mean over annulus
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r <= r2 && r >= r1 && image[j*axis_x + i]!=0){//check to see if we're within the annular patch
		    m[0] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[0] /= count;
	}//end if

  }
			
  if(strncmp(&stokes,"x",1)==0){//read in error I map
	if(fits_movabs_hdu(in,9,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading error I image from combine.fits",status);
	//set to zero
	set_to_zero(image,axis_x*axis_y);	
  }

  if(strncmp(&stokes,"q",1)==0){//read in q map
	if(fits_movabs_hdu(in,4,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading q image from combine.fits",status);
	//set to zero
	set_to_zero(image,axis_x*axis_y);
	if(dc=='y'){//compute mean over circular patch 8pix in diameter
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r<=4 && image[j*axis_x + i]!=0){//check to see if we're within the circular patch
		    m[1] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[1] /= count;
	}//end if
	if(dc2=='y'){//compute mean over annulus
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r <= r2 && r >= r1 && image[j*axis_x + i]!=0){//check to see if we're within the annular patch
		    m[1] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[1] /= count;
	}//end if
	
  }

  if(strncmp(&stokes,"y",1)==0){//read in error q map
	if(fits_movabs_hdu(in,5,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading error q image from combine.fits",status);	
	//set to zero
	set_to_zero(image,axis_x*axis_y);
  }
			
  if(strncmp(&stokes,"u",1)==0){//read in u map
	if(fits_movabs_hdu(in,6,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading u image from combine.fits",status);
	//set to zero
	set_to_zero(image,axis_x*axis_y);
	if(dc=='y'){//compute mean over circular patch 8pix in diameter
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r<=4 && image[j*axis_x + i]!=0){//check to see if we're within the circular patch
		    m[2] += image[j*axis_y + i];
		    count++;
		  }	
		}//end for loop
	  m[2] /= count;
	}//end if	
	if(dc2=='y'){//compute mean over annulus
	  count=0;
	  for(j=0;j<axis_y;j++)
	    for(i=0;i<axis_x;i++){
		  r = sqrt( pow(i-(axis_x-1)/2-0.5,2) + pow(j-(axis_y-1)/2-0.5,2) ); //compute radial distance
		  if(r <= r2 && r >= r1 && image[j*axis_x + i]!=0){//check to see if we're within the annular patch
		    m[2] += image[j*axis_x + i];
		    count++;
		  }	
		}//end for loop
	  m[2] /= count;
	}//end if
							
  }

  if(strncmp(&stokes,"z",1)==0){//read in error u map
	if(fits_movabs_hdu(in,7,0,&status))
	  Abort("FITS error with changing HDU's",status);
	if(fits_read_img(in,TDOUBLE,1,axis_x*axis_y,&null,image,&null2,&status))
	  Abort("FITS error with reading error u image from combine.fits",status);
	//set to zero
	set_to_zero(image,axis_x*axis_y);	
  }
			
  //close FITS file
  if(fits_close_file(in,&status))
    Abort("FITS error in closing data file",status);	
		  
  //close <NAMES.LIST>
  fclose(fp);
  
}

//Fuction to read file <POINTS.LIST>.  Output array P[N][3], where N is the # of points, second dimension is
//for x, y, and r.
double** readPoints(int *N){
  //declare variables and open cfile
  FILE *fp;
  fp = fopen(cfile,"r");
  double  **P; 
  int i=0,j=0;
  char temp[20]; 

  //read cfile to count # of points
  while(fscanf(fp,"%s",temp)!=EOF){
	i++;
  }
  i /= 3;
  (*N) = i;
  printf("Number of points read in from <POINTS.LIST>: %d\n",i);
  P = allocate2D(i,3);

  //now read in all data
  rewind(fp);
  i=0;
  while(fscanf(fp,"%s",temp)!=EOF){
    
	if(j!=3){
	  P[i][j] = atof(temp);
	  j++;
	  }
	else if(j==3){
	  j=0;
	  i++;
	  P[i][j] = atof(temp);
	  j++;
	  }
	  
  }

  fclose(fp);
  return(P);

}

//Function that prints out <POINTS.LIST> file out to the command line.  For debugging.
void printPoints(double **P, int N){

  int i;
  
  for(i=0;i<N;i++)
    printf("x: %lf, y: %lf, r: %lf\n",P[i][0],P[i][1],P[i][2]);

}

//Function for calculating the mean and stdev chi^2 over a cropped region defined by P
void crop(double **P, int N, double *XI, double *XQ, double *XU, double *MI, double *MQ, double *MU, int axis_x, int axis_y){

  int i,x,y,ci=0,cq=0,cu=0;
  double *ix, *qx, *ux, r, *im, *qm, *um, ave, var;
  
  for(i=0;i<N;i++){//primary for loop, cycle through each point
    
	//initialize count
	ci = 0;
	cq = 0;
	cu = 0;
	
	//check to make sure point is within resonable range
	if(P[i][0]>axis_x || P[i][0] <= 0 || P[i][1]>axis_y || P[i][1] <= 0)
	  Abort("Positions are within an incorrect range in <POINTS.LIST>",0);
  
    //cycle through X^2 maps, counting the # of pixels to read
	for(y=0;y<axis_y;y++)
	  for(x=0;x<axis_x;x++){
	    r = sqrt( pow((x-P[i][0]+1),2) + pow((y-P[i][1]+1),2) );//additional +1 is because the user defn's points starting at (1,1)
	    if(r<=P[i][2] && MI[y*axis_x+x]!=0)//check if pixel is in range for XI
		  ci++;
		if(r<=P[i][2] && MQ[y*axis_x+x]!=0)//check if pixel is in range for XQ
		  cq++;  
		if(r<=P[i][2] && MU[y*axis_x+x]!=0)//check if pixel is in range for XU
		  cu++;  
	  }
	
	//allocate memory
	ix = calloc(ci,sizeof(double));
	qx = calloc(cq,sizeof(double));
	ux = calloc(cu,sizeof(double));
	im = calloc(ci,sizeof(double));
	qm = calloc(cq,sizeof(double));
	um = calloc(cu,sizeof(double));
	
	//re-initialize counters
	ci = cq = cu = 0;
	
	//now cycle again through the data and read in info
	for(y=0;y<axis_y;y++)
	  for(x=0;x<axis_x;x++){
	    r = sqrt( pow((x-P[i][0]+1),2) + pow((y-P[i][1]+1),2) );
	    if(r<=P[i][2] && MI[y*axis_x+x]!=0){//check if pixel is in range for XI
		  ix[ci] = XI[y*axis_x+x];
		  im[ci] = 1;
		  ci++;
		  }
		if(r<=P[i][2] && MQ[y*axis_x+x]!=0){//check if pixel is in range for XQ
		  qx[cq] = XQ[y*axis_x+x];
		  qm[cq] = 1;
		  cq++;  
		  }
		if(r<=P[i][2] && MU[y*axis_x+x]!=0){//check if pixel is in range for XU
		  ux[cu] = XU[y*axis_x+x];
		  um[cu] = 1;
		  cu++;  
		  }
	  }

    //now compute average and stdev
	printf("For point %d:\n",i+1);
	avevar(ix,im,ci,&ave,&var);
	printf("Reduced Chi Squared mean and standard dev. for the I map: %lf, %lf\n",ave,sqrt(var));
	avevar(qx,qm,cq,&ave,&var);
	printf("Reduced Chi Squared mean and standard dev. for the Q map: %lf, %lf\n",ave,sqrt(var));
    avevar(ux,um,cu,&ave,&var);
	printf("Reduced Chi Squared mean and standard dev. for the U map: %lf, %lf\n",ave,sqrt(var));
	
	//free memory
	free(ix);
	free(qx);
	free(ux);
	free(im);
	free(qm);
	free(um);
	
  }//end of primary for loop

}

//Function that is used to inflate/reduce errors according to X^2 results
void correct_error_maps(char *file, double *sI, double *sQ, double *sU, double *XI, double *XQ, double *XU, double *MI, double *MQ, double *MU, int axis_x, int axis_y){
  int i,j,status=0;
  fitsfile *fp;

  printf("Correcting error maps upon request\n");

  //cycle through error maps 
  for(j=0;j<axis_y;j++)
    for(i=0;i<axis_x;i++){
	  //if statements for ensuring we aren't trying to deal with NAN's
	  if(MI[j*axis_x+i]!=0 && XI[j*axis_x+i]>=1)
	    sI[j*axis_x+i] = sqrt(XI[j*axis_x+i])*sI[j*axis_x+i];
	  if(MQ[j*axis_x+i]!=0 && XQ[j*axis_x+i]>=1)
	    sQ[j*axis_x+i] = sqrt(XQ[j*axis_x+i])*sQ[j*axis_x+i];
	  if(MU[j*axis_x+i]!=0 && XU[j*axis_x+i]>=1)
	    sU[j*axis_x+i] = sqrt(XU[j*axis_x+i])*sU[j*axis_x+i];
		
	  //if chi2 maps don't cover non-NAN area, inflate errrors *1000
	  if(MI[j*axis_x+i]==0)
	    sI[j*axis_x+i] = 1000*sI[j*axis_x+i];	
	  if(MQ[j*axis_x+i]==0)
	    sQ[j*axis_x+i] = 1000*sQ[j*axis_x+i];	
	  if(MU[j*axis_x+i]==0)
	    sU[j*axis_x+i] = 1000*sU[j*axis_x+i];		
				
	  //check if pixels = 0, if so assign to NAN
	  if(sI[j*axis_x+i]==0)
	    sI[j*axis_x+i] = NAN;
	  if(sQ[j*axis_x+i]==0)
	    sQ[j*axis_x+i] = NAN;
	  if(sU[j*axis_x+i]==0)
	    sU[j*axis_x+i] = NAN;			
	}

  //Open up cumulative FITS image for corrections    
  if(fits_open_file(&fp,file,READWRITE,&status))
	Abort("FITS error with opening sharp_combine output",status);
  if(fits_movabs_hdu(fp,9,0,&status))
    Abort("FITS error with changing HDU's",status); 
  if(fits_write_img(fp,TDOUBLE,1,axis_x*axis_y,sI,&status))
    Abort("FITS error in writing I error image",status);
  if(fits_movabs_hdu(fp,5,0,&status))
    Abort("FITS error with changing HDU's",status); 
  if(fits_write_img(fp,TDOUBLE,1,axis_x*axis_y,sQ,&status))
    Abort("FITS error in writing Q error image",status);
  if(fits_movabs_hdu(fp,7,0,&status))
    Abort("FITS error with changing HDU's",status); 
  if(fits_write_img(fp,TDOUBLE,1,axis_x*axis_y,sU,&status))
    Abort("FITS error in writing U error image",status);	
  if(fits_movabs_hdu(fp,1,0,&status))
    Abort("FITS error with changing HDU's",status);
  if(fits_write_key(fp,TSTRING,"CORRECTION","data corrected using chi2.c v2.4","",&status))
    Abort("FITS error in writing keyword",status);
  if(fits_close_file(fp,&status))
    Abort("FITS error in closing sharp_combine output",status);

}

//Function used for printing out an image of the I/sig_I ratio.  For debugging.
void I_over_sigI_ratio(int axis_x, int axis_y, double *I, double *sI){

  int x,y,status=0;
  double *out;
  fitsfile *fp;
  long naxes[2] = {axis_x,axis_y};
  
  //allocate memory
  out = calloc(axis_x*axis_y,sizeof(double));

  //cycle through image evaluating ratio
  for(y=0;y<axis_y;y++)
    for(x=0;x<axis_x;x++)
	  out[y*axis_x+x] = I[y*axis_x+x]/sI[y*axis_x+x];
	  
  //create output image and print	 
  if(fits_create_file(&fp,"!ratio.fits",&status))
    Abort("FITS error in generating output file",status);
  if(fits_create_img(fp,DOUBLE_IMG,2,naxes,&status))
    Abort("FITS error in generating image template",status);
  if(fits_write_key(fp,TSTRING,"EXTNAME","I_vs_sigI","Identification of HDU",&status))
    Abort("FITS error in writing ID key for HDU",status);
  if(fits_write_img(fp,TDOUBLE,1,axis_x*axis_y,out,&status))
    Abort("FITS error in writing image",status); 
  if(fits_close_file(fp,&status))
    Abort("FITS error in closing output file",status);	

  //free memory
  free(out);

}

//Function to read in and check image dimensions
void check_dimensions(int *axis_x, int *axis_y, char *Names){

  int Nodim,status=0,hdunum,bitpix,max=2;
  fitsfile *fp;
  long naxes[2];

  //read in all necessary info
  if(fits_open_file(&fp,Names,READONLY,&status))
    Abort("FITS error in opening file",status);
  if(fits_movabs_hdu(fp,2,0,&status))
    Abort("FITS error with changing HDU's",status);	
  if(fits_get_num_hdus(fp,&hdunum,&status))
    Abort("FITS error in reading # of HDU's in input file",status);
  if(fits_get_img_param(fp,max,&bitpix,&Nodim,naxes,&status))
    Abort("FITS error in reading image parameters",status);
  if(fits_close_file(fp,&status))
    Abort("FITS error in closing file",status);
	
  //do check	
  if(Nodim!=2){
    printf("Number of dimensions read: %d\n",Nodim);
    Abort("Number of dimensions for input file not equal to 2",status);
    }  
  if((*axis_x)==0 && (*axis_y)==0){
    (*axis_x) = (int)naxes[0];
    (*axis_y) = (int)naxes[1];
  }    	
  else{
    if( (*axis_x)!=(int)naxes[0] || (*axis_y)!=(int)naxes[1] )
      Abort("Image dimensions amoung data set vary",status);
  }
  if(hdunum==10 && v4=='y'){
    printf("Number of HDU's detected: %d, consistent with sharp_combine_v5 output",hdunum);
	Abort("You have specified the input as being from sharp_combine v4.xx but the input data looks like it's from sharp_combine_v5",status);
    }
  else if(hdunum==9 && v4=='n'){
    printf("Number of HDU's detected: %d, consistent with sharp_combine_v4.xx output",hdunum);
	Abort("You have specified the input as being from sharp_combine_v5 but the input data looks like it's from sharp_combine_v4.xx",status);
	}	

  //assign image dimensions			
  (*axis_x) = (int)naxes[0];			
  (*axis_y) = (int)naxes[1];		

}

//Function to compute the dc offset from info read in via read_bin(2)
void compute_dc(double **m, int b){

  int i,j;
  
  printf("PRINTING DC OFFSETS:\n");
  for(i=0;i<(b-1);i++)
    for(j=0;j<3;j++){
      m[i][j] = m[i][j] - m[b-1][j];//compute dc offset
	  if(j==0)
	    printf("I map DC offset for bin #%d: %lf\n",i+1,m[i][0]);
	  else if(j==1)	
	    printf("Q map DC offset for bin #%d: %lf\n",i+1,m[i][1]);
	  else if(j==2)
	    printf("U map DC offset for bin #%d: %lf\n",i+1,m[i][2]);
	
	}
  	  
}

//Function to compute inflation factor 
void compute_maxsig_factor(double *I, double *Q, double *U, double *sI, double *sQ, double *sU, double *XI, double *XQ, double *XU, double *M, int axis_x, int axis_y, double **P, int N){

  double p=0,pp=0,r;
  int i,x,y;
  
  for(i=0;i<axis_x*axis_y;i++){
    if(M[i]!=0){
      p += ( (pow(Q[i],2)+pow(U[i],2))/pow(I[i],4) )*pow(sI[i],2) + ( (pow(Q[i],2)/pow(I[i],2))/(pow(Q[i],2)+pow(U[i],2)) )*pow(sQ[i],2);
	  p += ( (pow(U[i],2)/pow(I[i],2))/(pow(Q[i],2)+pow(U[i],2)) )*pow(sU[i],2);
	  pp += ( (pow(Q[i],2)+pow(U[i],2))/pow(I[i],4) )*XI[i]*pow(sI[i],2) + ( (pow(Q[i],2)/pow(I[i],2))/(pow(Q[i],2)+pow(U[i],2)) )*XQ[i]*pow(sQ[i],2);
	  pp += ( (pow(U[i],2)/pow(I[i],2))/(pow(Q[i],2)+pow(U[i],2)) )*XU[i]*pow(sU[i],2);
	}
  }

  p /= (axis_x*axis_y);
  pp /= (axis_x*axis_y);
  
  printf("The inflation factor averaged over the map: %lf\n",sqrt(pp/p));

  if(points=='y'){
  
  for(i=0;i<N;i++){//primary for loop, cycle through each point
	
	//re-initialize vars
	p = 0;
	pp = 0;
	
	//check to make sure point is within resonable range
	if(P[i][0]>axis_x || P[i][0] <= 0 || P[i][1]>axis_y || P[i][1] <= 0)
	  Abort("Positions are within an incorrect range in <POINTS.LIST>",0);
  
    //work out size of memory to allocate
	for(y=0;y<axis_y;y++)
	  for(x=0;x<axis_x;x++){
	    r = sqrt( pow((x-P[i][0]+1),2) + pow((y-P[i][1]+1),2) );//additional +1 is because the user defn's points starting at (1,1)
	    if(r<=P[i][2] && M[y*axis_x+x]!=0){
		  p += ( (pow(Q[y*axis_x+x],2)+pow(U[y*axis_x+x],2))/pow(I[y*axis_x+x],4) )*pow(sI[y*axis_x+x],2);
		  p += ( (pow(Q[y*axis_x+x],2)/pow(I[y*axis_x+x],2))/(pow(Q[y*axis_x+x],2)+pow(U[y*axis_x+x],2)) )*pow(sQ[y*axis_x+x],2);
	      p += ( (pow(U[y*axis_x+x],2)/pow(I[y*axis_x+x],2))/(pow(Q[y*axis_x+x],2)+pow(U[y*axis_x+x],2)) )*pow(sU[y*axis_x+x],2);
	      pp += ( (pow(Q[y*axis_x+x],2)+pow(U[y*axis_x+x],2))/pow(I[y*axis_x+x],4) )*XI[y*axis_x+x]*pow(sI[y*axis_x+x],2);
		  pp += ( (pow(Q[y*axis_x+x],2)/pow(I[y*axis_x+x],2))/(pow(Q[y*axis_x+x],2)+pow(U[y*axis_x+x],2)) )*XQ[y*axis_x+x]*pow(sQ[y*axis_x+x],2);
	      pp += ( (pow(U[y*axis_x+x],2)/pow(I[y*axis_x+x],2))/(pow(Q[y*axis_x+x],2)+pow(U[y*axis_x+x],2)) )*XU[y*axis_x+x]*pow(sU[y*axis_x+x],2);  
	    }
	  
	  }
	  
	p /= (axis_x*axis_y);
    pp /= (axis_x*axis_y);

    printf("The inflation factor for point (%lf,%lf) is: %lf\n",P[i][0],P[i][1],sqrt(pp/p));

    }//end of primary for loop
  
  }

}



